{-
Module for loading and saving 24 bit Bitmaps
Bitmap specification taken from wikipedia http://en.wikipedia.org/wiki/BMP_file_format
Using the Binary.Get and Binary.Put portable libraries
Ideas taken from numerous tutorials over the Internet for Data.Binary.Get, Put and Data.ByteString
-}

module Extra.Bitmaps (readBmp, saveBmp) where

import Extra.Data.Binary.Get
import Extra.Data.Binary.Put
import qualified Data.ByteString.Lazy as BL
import Extra.Types
import Data.Word
import Char (ord)

-- main function --
readBmp :: FilePath -> IO (Bitmap)
readBmp fp = do list <- BL.readFile fp
                return(runGet parseBmp list)

saveBmp :: FilePath -> Bitmap -> IO ()
saveBmp fp bm = do str <- return(runPut (writefile bm))
                   BL.writeFile fp str

-- bitmap headers --

data BitmapHeader =
	BitmapHeader
	{
		bMagicNumber :: Int,
		bSize :: Int,
		bUnused :: Int,
		bUnused' :: Int,
		bOffset :: Int
	}

data ImageHeader =
	ImageHeader
	{
		iSize :: Int,
		iWidth :: Int,
		iHeight :: Int,
		iPlanes :: Int,
		iBitspp :: Int,
		iCompression :: Int,
		iImageSize :: Int,
		iPpmx :: Int,
		iPpmy :: Int,
		iPalette :: Int,
		iImpClrs :: Int
	}
-- helper functions --

readFileHeader :: Get BitmapHeader
readFileHeader = do
		bMagicNumber <- getWord16le
		bSize <- getWord32le
		bUnused <- getWord16le
		bUnused' <- getWord16le
		bOffset <- getWord32le
		return BitmapHeader
			{
				bMagicNumber = fromIntegral bMagicNumber,
				bSize = fromIntegral bSize,
				bUnused = fromIntegral bUnused,
				bUnused' = fromIntegral bUnused',
				bOffset = fromIntegral bOffset
			}

readImageHeader :: Get ImageHeader
readImageHeader = do
		iSize <- getWord32le
		iWidth <- getWord32le
		iHeight <- getWord32le
		iPlanes <- getWord16le
		iBitspp <- getWord16le
		iCompression <- getWord32le
		iImageSize <- getWord32le
		iPpmx <- getWord32le
		iPpmy <- getWord32le
		iPalette <- getWord32le
		iImpClrs <- getWord32le
		return ImageHeader
			{
				iSize = fromIntegral iSize,
				iWidth = fromIntegral iWidth,
				iHeight = fromIntegral iHeight,
				iPlanes = fromIntegral iPlanes,
				iBitspp = fromIntegral iBitspp,
				iCompression = fromIntegral iCompression,
				iImageSize = fromIntegral iImageSize,
				iPpmx = fromIntegral iPpmx,
				iPpmy = fromIntegral iPpmy,
				iPalette = fromIntegral iPalette,
				iImpClrs = fromIntegral iImpClrs
			}

readPixels :: Int -> Int -> Int -> Get [Tcolor] -- inspired by Daniel Renshaw's project from the last year
readPixels _ _ 0 = return []
readPixels wi 0 y = do
	skip (wi - (4 * (wi `div` 4)))
	readPixels wi wi (y - 1)
readPixels wi x y = do
	remaining <- remaining
	b <- getWord8
	g <- getWord8
	r <- getWord8
	rem <- readPixels wi (x - 1) y
	return ((round . fromIntegral $ r, round . fromIntegral $ g, round . fromIntegral $ b)  : rem)

colorsToBitmap :: Int -> [Tcolor] -> [[Tcolor]]
colorsToBitmap _ [] = []
colorsToBitmap w dat = [take w dat] ++ colorsToBitmap w (drop w dat)

parseBmp :: Get Bitmap
parseBmp = do
	headf <- readFileHeader
	headi <- readImageHeader
	if iBitspp headi == 24
		then skip (bOffset headf - 54)
		else fail "Unsupported bitmap type. Use 24 bit bitmaps"
	buff <- readPixels (iWidth headi) (iWidth headi) (iHeight headi)
	bitm <- return (colorsToBitmap (iWidth headi) buff)
	return (Bmp (iWidth headi) (iHeight headi) bitm)

-- bitmap saver

toWord8 :: Int -> Word8
toWord8 x = fromIntegral (x :: Int) :: Word8

toWord32 :: Int -> Word32
toWord32 x = fromIntegral (x :: Int) :: Word32

toWord16 :: Int -> Word16
toWord16 x = fromIntegral (x :: Int) :: Word16

writefile :: Bitmap -> Put
writefile (Bmp w h dat) = do 
 pad <- return (w - (4 * (w `div` 4)))
 size <- return (w*h*3 + pad*h)
 putWord8 (toWord8 (ord 'B'))
 putWord8 (toWord8 (ord 'M'))
 putWord32le (toWord32 (54+size))
 putWord16le (0 :: Word16)
 putWord16le (0 :: Word16)
 putWord32le (54 :: Word32)
 putWord32le (40 :: Word32)
 putWord32le (toWord32 w)
 putWord32le (toWord32 h)
 putWord16le (1 :: Word16)
 putWord16le (24 :: Word16)
 putWord32le (0 :: Word32)
 putWord32le (toWord32 size)
 putWord32le (2835 :: Word32)
 putWord32le (2835 :: Word32)
 putWord32le (0 :: Word32)
 putWord32le (0 :: Word32)
 writedata pad dat

writedata :: Int -> [[Tcolor]] -> Put
writedata _ [] = return ()
writedata pad (x:xs) = do 
 writerow x
 if pad == 1
  then putWord8 (0 :: Word8)
  else if pad == 2
        then putWord8 (0 :: Word8) >> putWord8 (0 :: Word8)
        else if pad == 3
             then putWord8 (0 :: Word8) >> putWord8 (0 :: Word8) >> putWord8 (0 :: Word8)
             else return ()
 writedata pad xs

writerow :: [Tcolor] -> Put
writerow [] = return ()
writerow ((r,g,b):xs) = do
 putWord8 (toWord8 b)
 putWord8 (toWord8 g)
 putWord8 (toWord8 r)
 writerow xs