module Mosaic
where

{-
Author: Martin Tsvetomirov Marinov
website: martinmarinov.info

Use the following commands:
main - creates the image. Reads from the database file (so the db file must exist).
analyze - generates a database file from the predefined folder for tiles and saves it to the harddisk
To assign a database file, tiles folder, etc. edit settings.hs with your favourite text editor and follow the instructions

What does the app do? It takes an image and for every given pixel it finds the "tile" with the closest average colour in the
palette to that pixel. Then it "equalizes" that tile so it's average colour is a perfect match with the color of the pixel
and then puts that tile onto the "canvas" and then saves the image to the hard disk.

If you want to show the file to the screen instead of saving it to a file, you could use MosaicDisplay.hs.
It won't work on some machines. Read the comment in the beginning of the file for more info.

The only supported input and output image format is BMP

If you manipulate big files, you should use the GHC command +RTS -K100M or if you are on a Windows machine execute "Run.bat"
-}

import Extra.Types
import Extra.Bitmaps
import Settings
import Extra.Trees
import Directory 
import Data.List
import Monad
import Array
import Random

main :: IO()
main = do putStr ("Opening " ++ file_name ++"\n")
          x <- readBmp file_name
          putStr ("Reading the database\n")
          datab <- readDB
          (Bmp w h _) <- return (x)
          tiles_in_height <- return(floor((fromIntegral h)*(fromIntegral tiles_in_width)*(fromIntegral tile_width)/((fromIntegral w)*(fromIntegral tile_height))))
          buffe1 <- return(resize (fromIntegral tiles_in_width) (fromIntegral tiles_in_height) x)
          putStr ("Generating...")
          buffe2 <- genmosaic w h tiles_in_height buffe1 datab
          putStr ("\nResizing output image\n")
          buffe <- return(combine x (resize w h buffe2))
          putStr ("Saving to File (this can take a while)...\n")
          saveBmp output_name buffe
          putStr ("Done\n")

-- mosaic generator

genmosaic :: Int -> Int -> Int -> Bitmap -> Record -> IO Bitmap
genmosaic w h tih (Bmp _ _ []) _ = return (Bmp 0 0 [])
genmosaic w h tih (Bmp _ _ (x:xs)) reco =
 do putChar '.'
    nx <- genrow w h tih x reco
    more <- genmosaic w h tih (Bmp 0 0 xs) reco
    return (above nx more)

genrow :: Int -> Int -> Int -> [Tcolor] -> Record -> IO Bitmap
genrow w h tih [] _ = return (Bmp 0 0 [])
genrow w h tih (x:xs) reco = do rnd <- randomRIO ((fromIntegral (-randomization)),(fromIntegral randomization))
                                (col, fp) <- return (get ((colorToInt x)+rnd*1000000) reco)
                                bmm <- readBmp fp
                                eqb <- return (resize (ceiling ((fromIntegral w)/(fromIntegral tiles_in_width))) (ceiling ((fromIntegral h)/(fromIntegral tih))) (equalize col x bmm))
                                xn <- genrow w h tih (xs) reco
                                return (nextto eqb xn)

nextto :: Bitmap -> Bitmap -> Bitmap -- equal size!
nextto x (Bmp _ _ []) = x
nextto (Bmp w h dat1) (Bmp w2 _ dat2) = Bmp (w+w2) h [ (dat1!!i) ++ (dat2!!i)| i <- [0..(h-1)]]

above :: Bitmap -> Bitmap -> Bitmap -- equal size!
above x (Bmp _ _ []) = x
above (Bmp w h dat1) (Bmp _ h2 dat2) = Bmp w (h+h2) (dat1 ++ dat2)

colorToInt :: Tcolor -> Int
colorToInt (r, g, b) = 256*256*r+256*g+b


equalize :: Int -> Tcolor -> Bitmap -> Bitmap -- the first Int color is read from the db
equalize f (r2, g2, b2) (Bmp w h dat) = Bmp w h [[(eq (rr+r),eq (gg+g),eq (bb+b)) | (rr, gg, bb) <- dsm] | dsm <-dat]
 where (r1, g1, b1) = intToColor f
       r = r2 - r1
       g = g2 - g1
       b = b2 - b1
       eq s
        | s < 0 = 0
        | s > 255 = 255
        | otherwise = s

intToColor :: Int -> Tcolor
intToColor i = (r, g, b)
 where b  = i `mod` 256
       br = i `div` 256
       g  = br `mod` 256
       r  = br `div` 256

-- database generation and loaders. based on Tutorial 6

type Record = Keymap Int FilePath -- this is a record in the database. It contains the filepath to a given file and its `avarage` colour

readDB :: IO Record
readDB = do db <- readFile db_name
            return $ fromList $ map readLine $ lines db

readLine :: String -> (Int, FilePath)
readLine str = (sti (c), a)
    where
      (c,a) = splitUpon ',' str

splitUpon :: Char -> String -> (String,String)
splitUpon _ "" = ("","")
splitUpon c (x:xs) | x == c    = ("",xs)
                   | otherwise = (x:ys,zs)
                   where
                     (ys,zs) = splitUpon c xs

sti :: String -> Int -- StrinToInt conversation
sti st = read st::Int

saveDB :: [(Int, FilePath)] -> IO ()
saveDB lst = writeFile db_name (generate lst)
 where generate [] = ""
       generate ((cl,s):xs) = show cl ++ "," ++ s ++ "\n" ++ generate xs

-- picture manipulation
getfiles :: IO ([FilePath]) -- gets .bmp file names in a given directory
getfiles = do rawfiles <- getDirectoryContents tiles_folder
              return (map ((++) (tiles_folder ++ "/")) (filter (isSuffixOf ".bmp") rawfiles))

averagecol :: IO (Bitmap) -> IO (Int)
averagecol s = let
    rs :: [Tcolor] -> Int
    rs lst = round (fromIntegral (sum [r | (r, _, _) <- lst]) / (fromIntegral.length $ lst))
    gs :: [Tcolor] -> Int
    gs lst = round (fromIntegral (sum [g | (_, g, _) <- lst]) / (fromIntegral.length $ lst))
    bs :: [Tcolor] -> Int
    bs lst = round (fromIntegral (sum [b | (_, _, b) <- lst]) / (fromIntegral.length $ lst))
    getpixdata (Bmp _ _ dat) = concat dat
    in do 
        x <- s
        dat <- return (getpixdata x)
        putStr (".")
        return(colorToInt (rs dat, gs dat, bs dat))

analyze :: IO ()
analyze = do fls <- getfiles
             putStr ("Analyzing")
             db <- mapM (averagecol . readBmp) fls
             tmp <- return (zip db fls)
             putStr ("\nSaving data to file. Please, wait (this can take a while)...\n")
             saveDB tmp
             putStr ("Database file " ++ db_name ++ " saved\n")

resize :: Int -> Int -> Bitmap -> Bitmap -- implementation of the "nearest neighbour" alghorithm for resizing images
resize nw nh (Bmp w h bm) = Bmp nw nh [[cviat i j | i <- [1..nw]] | j <- [1..nh]]
 where pixc = w * h
       ar = array (1,pixc) (zip [1..pixc] (concat bm))
       kx = (fromIntegral nw)/(fromIntegral w)
       ky = (fromIntegral nh)/(fromIntegral h)
       cdx = (fromIntegral nw)/2 -- center point of the dest image
       cdy = (fromIntegral nh)/2 -- center point of the dest image
       cox = (fromIntegral w)/2 -- center point of the orig image
       coy = (fromIntegral h)/2 -- center point of the orig image
       cviat i j = ar ! arind
        where udx = (fromIntegral i) - cdx -- "unreal" dest coordinates
              udy = (fromIntegral j) - cdy
              uox = udx/kx -- "unreal" orig coordinates
              uoy = udy/ky
              ox = round (uox + cox)
              oy = round (uoy + coy)
              arindt = (oy - 1) * w + ox
              arind
               | arindt < 1 = 1
               | arindt > pixc = pixc
               | otherwise = arindt

combine :: Bitmap -> Bitmap -> Bitmap
combine (Bmp h w one) (Bmp _ _ two) = (Bmp h w (mergerow one two))
 where mergerow [] _ = []
       mergerow _ [] = []
       mergerow (x:xs) (y:ys) = (mergecol x y) : (mergerow xs ys)
        where mergecol [] _ = []
              mergecol _ [] = []
              mergecol (xx:xxs) (yy:yys) = (mcol xx yy) : (mergecol xxs yys)
               where mcol (r1, g1, b1) (r2, g2, b2) = (round(rr1 * k + rr2 * (1-k)), round(gr1 * k + gr2 * (1-k)), round(br1 * k + br2 * (1-k)))
                      where rr1 = fromIntegral r1
                            gr1 = fromIntegral g1
                            br1 = fromIntegral b1
                            rr2 = fromIntegral r2
                            gr2 = fromIntegral g2
                            br2 = fromIntegral b2
                            k   = blending_per/100